#######################################################################################################################################################################################################
#
# Changes to this type
# General
#     2021-Sep-30: add abstract type for soil vulnerability curve
#
#######################################################################################################################################################################################################
"""

$(TYPEDEF)

Hierarchy of AbstractSoilVC:
- [`BrooksCorey`](@ref)
- [`VanGenuchten`](@ref)
"""
abstract type AbstractSoilVC{FT<:AbstractFloat} end


#######################################################################################################################################################################################################
#
# Changes to this structure
# General
#     2021-Sep-30: define this structure with no default constructor
#     2022-May-25: fix documentation
#
#######################################################################################################################################################################################################
"""

$(TYPEDEF)

Brooks Corey soil parameters

# Fields

$(TYPEDFIELDS)

"""
struct BrooksCorey{FT<:AbstractFloat} <:AbstractSoilVC{FT}
    # parameters that do not change with time
    "Soil type"
    TYPE::String
    "Soil b"
    B::FT
    "Potential at saturation `[MPa]`"
    Ψ_SAT::FT
    "Saturated soil volumetric water content"
    Θ_SAT::FT
    "Residual soil volumetric water content"
    Θ_RES::FT
end


#######################################################################################################################################################################################################
#
# Changes to this structure
# General
#     2021-Sep-30: define this structure with two default constructors from an incomplete parameter set
#     2021-Nov-24: move the constructors out of the structure with examples
#     2022-May-25: fix documentation
#
#######################################################################################################################################################################################################
"""

$(TYPEDEF)

van Genuchten soil parameters

# Fields

$(TYPEDFIELDS)

"""
struct VanGenuchten{FT<:AbstractFloat} <:AbstractSoilVC{FT}
    # parameters that do not change with time
    "Soil type"
    TYPE::String
    "Soil α is related to the inverse of the air entry suction, α > 0"
    Α::FT
    "Soil n is Measure of the pore-size distribution"
    N::FT
    "Soil m = 1 - 1/n"
    M::FT
    "Saturated soil volumetric water content"
    Θ_SAT::FT
    "Residual soil volumetric water content"
    Θ_RES::FT
end


#######################################################################################################################################################################################################
#
# Changes to this constructor
# General
#     2021-Sep-30: add constructor function
#     2022-May-25: fix documentation
#
#######################################################################################################################################################################################################
"""

    VanGenuchten{FT}(name::String, α::Number, n::Number, θ_sat::Number, θ_res::Number) where {FT<:AbstractFloat}

Constructor for [`VanGenuchten`](@ref), given
- `name` Soil type name
- `α` van Genuchten α
- `n` van Genuchten n
- `θ_sat` Saturated soil volumetric water content
- `θ_res` Residual soil volumetric water content

---
# Examples
```julia
vg = VanGenuchten{FT}("Test", FT(100), FT(2), FT(0.5), FT(0.1));
```
"""
VanGenuchten{FT}(name::String, α::Number, n::Number, θ_sat::Number, θ_res::Number) where {FT<:AbstractFloat} = VanGenuchten{FT}(name, α, n, 1-1/n, θ_sat, θ_res)


#######################################################################################################################################################################################################
#
# Changes to this constructor
# General
#     2021-Sep-30: add constructor function
#     2022-May-25: fix documentation
#
#######################################################################################################################################################################################################
"""

    VanGenuchten{FT}(name::String) where {FT<:AbstractFloat}

Constructor for [`VanGenuchten`](@ref), given
- `name` Soil type name, need to be Sand, Loamy Sand, Sandy Loam, Loam, Sandy Clay Loam, Silt Loam, Silt (default), Clay Loam, Silty Clay Loam, Sandy Clay, Silty Clay, and Clay.

---
# Examples
```julia
vg = VanGenuchten{FT}("");
vg = VanGenuchten{FT}("Loam");
```
"""
VanGenuchten{FT}(name::String) where {FT<:AbstractFloat} = (
    # Parameters from Silt soil
    _p = [ 163.2656, 1.37, 0.46, 0.034];

    # switch name
    if name=="Sand"
        _p = [1479.5945, 2.68, 0.43, 0.045];
    elseif name=="Loamy Sand"
        _p = [1265.3084, 2.28, 0.41, 0.057];
    elseif name=="Sandy Loam"
        _p = [ 765.3075, 1.89, 0.41, 0.065];
    elseif name=="Loam"
        _p = [ 367.3476, 1.56, 0.43, 0.078];
    elseif name=="Sandy Clay Loam"
        _p = [ 602.0419, 1.48, 0.39, 0.100];
    elseif name=="Silt Loam"
        _p = [ 204.0820, 1.41, 0.45, 0.067];
    elseif name=="Silt"
        _p = [ 163.2656, 1.37, 0.46, 0.034];
    elseif name=="Clay Loam"
        _p = [ 193.8779, 1.31, 0.41, 0.095];
    elseif name=="Silty Clay Loam"
        _p = [ 102.0410, 1.23, 0.43, 0.089];
    elseif name== "Sandy Clay"
        _p = [ 275.5107, 1.23, 0.38, 0.100];
    elseif name=="Silty Clay"
        _p = [  51.0205, 1.09, 0.36, 0.070];
    elseif name=="Clay"
        _p = [  81.6328, 1.09, 0.38, 0.068];
    else
        @warn "Soil type $(name) not recognized, use Silt instead.";
        name = "Silt";
    end;

    # return a new struct
    return VanGenuchten{FT}(name, _p[1], _p[2], _p[3], _p[4])
);
